# Import necessary libraries
import subprocess
import os


# Subseqs (keep track of extracted scaffolds)
subSeqs = []

# Paths definitions to use
prot_path = './prot/'
genome_path = '/mnt/NFS_UPF/soft/genomes/2021/Paroedura_picta/genome.fa'
genIndex_path = '/mnt/NFS_UPF/soft/genomes/2021/Paroedura_picta/genome.index'



# Command functions
def tblastn(qname):
    return 'tblastn -query {}{}.fa -db {} -out ./RESULTS/{}/{}_blast -outfmt "6 qseqid sseqid evalue sstart send slen"'.format(
        prot_path, qname, genome_path, qname, qname)


def fastafetch(sname, qname):
    return 'fastafetch {} {} "{}" > ./RESULTS/fetch/{}.fa'.format(
        genome_path, genIndex_path, sname, sname)


def fastasubseq(sname, fasta_start, fasta_length, start, end, qname):
    return 'fastasubseq ./RESULTS/fetch/{}.fa {} {} > ./RESULTS/{}/{}/{}_{}_{}.fa'.format(
        sname, fasta_start, fasta_length, qname, sname, sname, start, end)


def exonerate(qname, sname, start, end):
    return 'exonerate -m p2g --showtargetgff -q ./prot/{}.fa -t ./RESULTS/{}/{}/{}_{}_{}.fa | egrep -w exon > ./RESULTS/{}/{}/{}.exonerate.gff'.format(
        qname, qname, sname, sname, start, end, qname, sname, sname)


def fastaseqfromGFF(qname, sname, start, end):
    return 'fastaseqfromGFF.pl ./RESULTS/{}/{}/{}_{}_{}.fa ./RESULTS/{}/{}/{}.exonerate.gff > ./RESULTS/{}/{}/{}_translate.fa'.format(
        qname, sname, sname, start, end, qname, sname, sname, qname, sname, sname)


def fastatranslate(qname, sname):
    return 'fastatranslate ./RESULTS/{}/{}/{}_translate.fa -F 1 | tr "*" "X" > ./RESULTS/{}/{}/{}_PRED.fa'.format(
        qname, sname, sname, qname, sname, sname)


def tcoffe(qname, sname):
    return 't_coffee ./prot/{}.fa ./RESULTS/{}/{}/{}_PRED.fa -outfile=./RESULTS/{}/{}/{}_ALIGNMENT'.format(
        qname, qname, sname, sname, qname, sname, sname)


# Aux functions
def shell(cmd):
    subprocess.run(cmd, shell=True)


def process_blast(qname):
    '''
    :param:  Name of query protein
    :return: Dictionary with the 'valid' scaffolds and their start, end position and scaffold length
    '''

    dict = {}
    with open('./RESULTS/{}/{}_blast'.format(qname, qname), 'r') as f:
        for row in f:
            # FORMAT: QNAME; SNAME; EVAL; START; END; SLEN
            line = row[:].split()
            sname = line[1]
            # Calculate start and end position in case forward/reverse seq
            start = min(int(line[3]), int(line[4]))
            end = max(int(line[3]), int(line[4]))
            eval = float(line[2])
            slen = int(line[5])

            if eval < 0.0001:
                if sname not in dict:
                    dict[sname] = {'start': start, 'end': end, 'slen': slen}
                else:
                    dict[sname] = {'start': min(start, dict[sname]['start']),
                                   'end': max(end, dict[sname]['end']),
                                   'slen': slen}
    f.close()
    return dict


def scaffold_to_txt(sname, start, end, slen, qname):
    '''
    :return: Saves to a txt the scaffold info
    '''
    return 'echo {} \t {} \t {} \t {} >> ./RESULTS/{}/summary.txt'.format(
        sname, start, end, slen, qname)


def valid_region(start, end, slen):
    '''
    :param: Scaffold start position, end position and length
    :return: Valid start position and length for performing fastasubseq
    '''

    len_hit = end - start
    start_fasta = 0
    len_fasta = 0

    # If scaffold too short, keep entire scaffold
    if slen <= 100000:
        len_fasta = slen

    # If we cannot subtract 50k from the start
    elif (start - 50000) < 0:
        len_fasta = start + len_hit + 50000

    # If we cannot add 50k from the end
    elif (end + 50000) > slen:
        start_fasta = start - 50000
        len_fasta = 50000 + len_hit + (slen - end)

    else:
        start_fasta = start - 50000
        len_fasta = 50000 + len_hit + 50000

    return start_fasta, len_fasta


if __name__ == "__main__":
	
	# Before executing the script make sure your query proteins
	# are in a directory ./prot/ , also check the export path

    # Create a directory for storing results
    if not os.path.exists('./RESULTS'):
        os.mkdir('./RESULTS')
        
    # Create a directory for storing fetched scaffolds
    if not os.path.exists('./RESULTS/fetch/'):
        os.mkdir('./RESULTS/fetch/')

    # Iterate over each protein query
    for protein in os.listdir(prot_path):
        protName = protein.replace('.fa', '')

        # Create a protein directory for storing results
        if not os.path.exists('./RESULTS/{}'.format(protName)):
            os.mkdir('./RESULTS/{}'.format(protName))

        # Perform TBLASTN
        shell(tblastn(protName))

        # Process results of TBLASTN
        scaffolds = process_blast(protName)

        for scaffold_name in scaffolds:
            sname = scaffold_name
            start = scaffolds[scaffold_name]['start']
            end = scaffolds[scaffold_name]['end']
            slen = scaffolds[scaffold_name]['slen']

            # Store scaffold info
            shell(scaffold_to_txt(sname, start, end, slen, protName))

            # Create a scaffold directory for storing results
            if not os.path.exists('./RESULTS/{}/{}/'.format(protName, sname)):
                os.mkdir('./RESULTS/{}/{}/'.format(protName, sname))


            # Pipeline
            # Avoid fetching the same scaffold again, since consumes lots of memory disk space
            if sname not in subSeqs:
                shell(fastafetch(sname, protName))
                subSeqs.append(sname)

            start_fasta, len_fasta = valid_region(start, end, slen)
            shell(fastasubseq(sname, start_fasta, len_fasta, start, end, protName))

            shell(exonerate(protName, sname, start, end))

            shell(fastaseqfromGFF(protName, sname, start, end))
            shell(fastatranslate(protName, sname))

            shell(tcoffe(protName, sname))


